/*    Copyright 2008 Wagner Monzyne
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##
##
##    mdcp - Multi-Copy Device
##    Author: Wagner Monzyne (MONZYNE, W.)
##    file: mdcp.cc
##
#########################################################################*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
//-----------------------------------------------------------

#define PROG_NAME "mdcp"
//-----------------------------------------------------------

typedef struct 
{
  unsigned char qtd;
  unsigned char qtdActive;
  int          *files;
  size_t       *w_size;
  char        **names;
} FILES;

#define NBUFFS 3

typedef struct 
{
  unsigned   long  len;
  unsigned   long  used[NBUFFS];
  unsigned   char *buff[NBUFFS];
             char  stat[NBUFFS];
  pthread_mutex_t  lock[NBUFFS];
  pthread_cond_t   sign[NBUFFS];
} BUFFER;

enum { BUF_UNUSED, BUF_BUSY, BUF_USED, BUF_READONLY };


int arqIn;
FILES arqsOut;

BUFFER buffer;

#define DEFAULT_BS 32768UL;
unsigned long bs;
unsigned long long count;
static unsigned long long r_size = 0;

//-----------------------------------------------------------

void iniciarFILES( int qtd )
{
  arqsOut.qtd       = qtd;
  arqsOut.qtdActive = qtd;
  arqsOut.files     = (int *)   malloc( qtd * sizeof( int    ) );
  arqsOut.w_size    = (size_t*) malloc( qtd * sizeof( size_t ) );
  arqsOut.names     = (char**)  malloc( qtd * sizeof( char*  ) );
  for ( int f = 0; f < qtd; arqsOut.w_size[f++] = 0 );
}
//-----------------------------------------------------------

void liberarFILES()
{
  arqsOut.qtd    = 0;
  free( arqsOut.files );
}
//-----------------------------------------------------------


void iniciarBuff()
{
  buffer.len = bs;
  for ( int b = 0; b < NBUFFS; b++ )
  {
    buffer.used[b] = 0;
    buffer.stat[b] = BUF_UNUSED;
    pthread_mutex_init( &buffer.lock[b], NULL );
    pthread_cond_init( &buffer.sign[b], NULL );
    buffer.buff[b] = (unsigned char*) malloc( buffer.len );
    memset( buffer.buff[b], 0, buffer.len );
  }
}
//-----------------------------------------------------------

void liberarBuff()
{
  buffer.len = 0;
  for ( int b = 0; b < NBUFFS; b++ )
  {
    buffer.used[b] = 0;
    buffer.stat[b] = BUF_UNUSED;
    pthread_mutex_destroy( &buffer.lock[b] );
    pthread_cond_destroy( &buffer.sign[b] );
    free( buffer.buff[b] );
    buffer.buff[b] = NULL;
  }
}
//-----------------------------------------------------------

#define NEXT_BUFF( b ) ( (b +1) % NBUFFS )
#define PREV_BUFF( b ) ( (b + NBUFFS-1) % NBUFFS )
#define  INC_BUFF( b ) ( b = NEXT_BUFF( b ) )
#define  DEC_BUFF( b ) ( b = PREV_BUFF( b ) )
//-----------------------------------------------------------

void *readFile( void * )
{
  unsigned char nb = 0;
  unsigned long count = 0;
  bool end = false;

  pthread_mutex_lock( &buffer.lock[nb] );
  buffer.stat[nb] = BUF_BUSY;
  pthread_mutex_unlock( &buffer.lock[nb] );

  buffer.used[nb] = read( arqIn, buffer.buff[nb], buffer.len );
  if( buffer.used < 0 )
  {
    buffer.used[nb] = 0;
    end = true;
  }
  else if ( buffer.used[nb] == 0 )
  {
    end = true;
  }
  else
  {
    r_size += buffer.used[nb];
  }

  ++count;

  pthread_mutex_lock( &buffer.lock[nb] );
  buffer.stat[nb] = BUF_USED;
  pthread_cond_broadcast( &buffer.sign[nb] );
  pthread_mutex_unlock( &buffer.lock[nb] );
  
  while ( (arqsOut.qtdActive > 0) && ( !end ) && ( !::count || ( count < ::count ) ) )
  {
    INC_BUFF( nb );

    // aguarda ate que o buffer esteja livre
    pthread_mutex_lock( &buffer.lock[nb] );
    while( buffer.stat[nb] != BUF_UNUSED )
    {
      pthread_cond_wait( &buffer.sign[nb], &buffer.lock[nb] );
    }

    buffer.stat[nb] = BUF_BUSY;
    pthread_mutex_unlock( &buffer.lock[nb] );

    buffer.used[nb] = read( arqIn, buffer.buff[nb], buffer.len );
    if( buffer.used[nb] < 0 )
    {
      buffer.used[nb] = 0;
      end = true;
    }
    else if ( buffer.used[nb] == 0 )
    {
      end = true;
    }
    else
    {
      r_size += buffer.used[nb];
    }

    ++count;

    pthread_mutex_lock( &buffer.lock[nb] );
    buffer.stat[nb] = BUF_USED;
    pthread_cond_broadcast( &buffer.sign[nb] );
    pthread_mutex_unlock( &buffer.lock[nb] );
  }

  if ( !end )
  {
  // preenche um último buffer com tamanho zero
  // para indicar fim da cópia
    INC_BUFF( nb );

    pthread_mutex_lock( &buffer.lock[nb] );
    while( buffer.stat[nb] != BUF_UNUSED  )
    {
      pthread_cond_wait( &buffer.sign[nb], &buffer.lock[nb] );
    }

    buffer.used[nb] = 0;
    buffer.stat[nb] = BUF_USED;  

    pthread_cond_broadcast( &buffer.sign[nb] );
    pthread_mutex_unlock( &buffer.lock[nb] );
  }
}
//-----------------------------------------------------------

pthread_cond_t  cond_allFilesReady = PTHREAD_COND_INITIALIZER;
pthread_mutex_t mute_nfilesDone    = PTHREAD_MUTEX_INITIALIZER;
//-----------------------------------------------------------

inline void destFileError( unsigned char nfile )
{
  fprintf(stderr, "%s: Error in destination: Maybe disk is full: \"%s\"\n", PROG_NAME,  arqsOut.names[nfile]);
  fprintf(stderr, "%s: Destination removed: \"%s\": %d bytes was written yet\n", PROG_NAME,  arqsOut.names[nfile], arqsOut.w_size[nfile]);

}
//-----------------------------------------------------------

void *writeFile( void *_nfile )
{
  static unsigned char nfilesDone = 0;
  unsigned char nfile = *(unsigned char*)_nfile;
  register unsigned char nb = 0;
  long long written;
  bool end   = false;
  bool error = false;

  while ( !end )
  {
    // aguarda ate que o buffer esteja pronto
    pthread_mutex_lock( &buffer.lock[nb] );
    while( buffer.stat[nb] != BUF_USED && (buffer.stat[nb] != BUF_READONLY) )
    {
      pthread_cond_wait( &buffer.sign[nb], &buffer.lock[nb] );
    }
    buffer.stat[nb] = BUF_READONLY;
    pthread_mutex_unlock( &buffer.lock[nb] );

    if ( buffer.used[nb] )
    {
      written = write( arqsOut.files[ nfile ], buffer.buff[nb], buffer.used[nb] );
      if ( written <= 0 )
      {
        destFileError( nfile );
        error = true;
        end   = true;
      }
      else
      {
        arqsOut.w_size[nfile] += written;
      }
    }
    else
    {
      end = true;
    }

    pthread_mutex_lock( &mute_nfilesDone );
    if ( (nfilesDone + 1) < arqsOut.qtdActive )
    {
      if ( !error )
      {
        ++nfilesDone;
      }
      else
      {
        --arqsOut.qtdActive;
      }
      pthread_cond_wait( &cond_allFilesReady, &mute_nfilesDone );
    }
    else
    {
      if ( error ) --arqsOut.qtdActive; // <-- it file fails. removed from work...

      if ( !arqsOut.qtdActive )
      { /* it is the last file to write, and fails...
        /* then, it have to "UNUSE" all the buffers because readFile
        /* need to leave the work as soon as possibile
        */
        unsigned char _nb = PREV_BUFF( nb );
        while ( nb != _nb )
        {
          // wait if buffer is busy
          pthread_mutex_lock( &buffer.lock[_nb] );
          while( buffer.stat[_nb] == BUF_BUSY )
          {
            pthread_cond_wait( &buffer.sign[_nb], &buffer.lock[_nb] );
          }

          buffer.used[_nb] = 0;
          buffer.stat[_nb] = BUF_UNUSED;

          pthread_cond_signal( &buffer.sign[_nb] );
          pthread_mutex_unlock( &buffer.lock[_nb] );
          DEC_BUFF( _nb );
        }
      }

      buffer.used[nb] = 0;

      pthread_mutex_lock( &buffer.lock[nb] );
      buffer.stat[nb] = BUF_UNUSED;
      pthread_cond_signal( &buffer.sign[nb] );
      pthread_mutex_unlock( &buffer.lock[nb] );

      nfilesDone = 0;
      pthread_cond_broadcast( &cond_allFilesReady );
    }
    pthread_mutex_unlock( &mute_nfilesDone );

    INC_BUFF( nb );
  }
}
//-----------------------------------------------------------

void transferir()
{
  r_size = 0;

  pthread_t threadRead, *threadsWrite;
  threadsWrite = (pthread_t*) malloc( arqsOut.qtd * sizeof( pthread_t ) );

  unsigned char *aux = (unsigned char*) malloc( arqsOut.qtd * sizeof( unsigned char ) ) ;
  for ( unsigned char nfile = 0; nfile < arqsOut.qtd; nfile++ )
  {
    aux[nfile] = nfile;
    pthread_create( &threadsWrite[nfile], NULL, &writeFile, (void*)&aux[nfile] );
  }
  pthread_create( &threadRead, NULL, &readFile, NULL );

  pthread_join( threadRead, NULL );
  for ( unsigned char nfile = 0; nfile < arqsOut.qtd; nfile++ )
  {
    pthread_join( threadsWrite[nfile], NULL );
  }

  free( threadsWrite );
  free( aux );
}
//-----------------------------------------------------------

struct timespec timeStarted;
double duracaoCopiaSecs = 0.0;
//-----------------------------------------------------------

inline void startTime()
{
  clock_gettime( NULL, &timeStarted );
}
//-----------------------------------------------------------

inline void markTime()
{
  struct timespec timeFinished;
  clock_gettime( NULL, &timeFinished );

  duracaoCopiaSecs = \
        (timeFinished.tv_sec - timeStarted.tv_sec) \
      + ( timeFinished.tv_nsec - timeStarted.tv_nsec)/(1000.0*1000*1000);
}
//-----------------------------------------------------------

void showStatistics()
{
  size_t w_total = 0;
  printf("Done.\n");
  printf("Bytes readed : %d\n", r_size );
  printf("Bytes written :");
  if ( arqsOut.qtd > 1 )
  {
    printf("\n");
    for ( unsigned char f = 0; f < arqsOut.qtd; f++ )
    {
      printf("\t%d @ %s\n", arqsOut.w_size[f], arqsOut.names[f] );
      w_total += arqsOut.w_size[f];
    }
    printf("\t%d TOTAL\n", w_total );
  }
  else
  {
    w_total = arqsOut.w_size[0];
    printf(" %d\n", w_total );
  }
  printf("Time elapsed : %0.4fsecs\n", duracaoCopiaSecs );
  printf("%sransfer rate : %0.fb/s\n\n",(arqsOut.qtd>1?"Equivalent t":"T"),
        w_total / duracaoCopiaSecs );
}
//-----------------------------------------------------------

void exitErrorUsage()
{
  printf("Usage:\n");
  printf("\t%s <bs> <count> <fileIn> <fileOut1> ... <fileOutN>\n", PROG_NAME );
  printf("\t<bs>\t\t: Block size. Leave it 0 to default size 32768\n");
  printf("\t<count>\t\t: Copy only <count> blocks. Leave it 0 for copy\n"\
         "\t\t\t  until find end of file\n");
  printf("\t<fileIn>\t: Device to be read\n");
  printf("\t<fileOutN>\t: Device to be written\n");
  printf("\n");

  exit( -1 );
}
//-----------------------------------------------------------

void fatalError()
{
  // TODO: clean the system...
  exit( -1 );
}
//-----------------------------------------------------------

int openSourceFile( char *name )
{
  arqIn = open( name, O_RDONLY, 0 );
  if ( !( arqIn > 0 ) )
  {
    fprintf(stderr, "%s: Fatal Error: Filed to open the SOURCE \"%s\"\n", PROG_NAME, name);
    return 0;
  }
  return arqIn;
}
//-----------------------------------------------------------

void closeSourceFile()
{
  close( arqIn );
}
//-----------------------------------------------------------

int openDestFile( unsigned char nfile, char *name )
{
  arqsOut.files[nfile] = open( name, O_WRONLY | O_TRUNC | O_CREAT, 0 );
  if ( !( arqsOut.files[nfile] > 0 ) )
  {
    fprintf(stderr, "%s: Fatal Error: Filed to open the DESTINATION \"%s\"\n", PROG_NAME, name);
    return 0;
  }
  arqsOut.names[nfile] = name;
  return arqsOut.files[nfile];
}
//-----------------------------------------------------------

void closeDestFiles()
{
  for( int nfout = 0; nfout < arqsOut.qtd; nfout++ )
  {
    if ( 0 < arqsOut.files[ nfout ] )
    {
      close( arqsOut.files[ nfout ] );
    }
  }
}
//-----------------------------------------------------------

int main( int argc, char *argv[] )
{
  enum { BS = 1, COUNT, ARQIN, ARQOUT };

  // no mínimo devem ser informados 3 argumentos
  if ( argc < 5 ) {
    exitErrorUsage();
  }


  // bs deve ser maior que zero, pois é o buffer de transferência
  bs    = atol( argv[BS] );
  if (bs < 0) {
    exitErrorUsage();
  }
  else if ( bs == 0 )
  {
    bs = DEFAULT_BS;
  }

  // count pode ser zero (até fim do arquivo) ou positivo
  count = atol( argv[COUNT] );
  if (count < 0) {
    exitErrorUsage();
  }

  if ( 0 == openSourceFile( argv[ARQIN] ) )
    fatalError();

  // try to open targets files
  iniciarFILES( argc - ARQOUT );
  for( int nfout = ARQOUT; nfout < argc; nfout++ )
  {
    if ( 0 == openDestFile( (nfout - ARQOUT), argv[nfout] ) )
    {
      closeSourceFile();
      closeDestFiles();
      liberarFILES();
      fatalError();
    }
  }

  iniciarBuff();

  startTime();

  transferir();

  markTime();
  showStatistics();

  liberarBuff();

  closeSourceFile();
  closeDestFiles();
  liberarFILES();

  return 0;
}
//-----------------------------------------------------------
